from unittest import TestCase
import pkg_resources
from random import randint
from dateutil.tz import tzutc
from datetime import datetime


def random_row(n_rows):
    return randint(0, n_rows-1)


class IoPandasTests(TestCase):
    """Acceptance and unit tests for loading pandas Datafram from json files.
    """

    def test_load_dataframe_from_json_file_1p_1t_ma(self):
        """Acceptance test for loading data from a Meadows
        .json file download containing data for a single MA task,
        single participant. Should have data and descriptors
        as found in file.
        """
        from meadows.io.pandas import df_from_tree_file
        fname = 'Meadows_myExp_v_v2_profound-mammoth_1_tree.json'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        df = df_from_tree_file(fpath)
        self.assertEqual(df['trial'].unique().size, 56)
        N_PLACEMENTS_TOTAL = TASK_ROWS = 205
        self.assertEqual(len(df), N_PLACEMENTS_TOTAL)
        ## file-level descriptors
        self.assertEqual(df['participant'][random_row(TASK_ROWS)], 'profound-mammoth')
        self.assertEqual(df['task_index'][random_row(TASK_ROWS)], 1)
        self.assertEqual(df['task_name'][random_row(TASK_ROWS)], 'multiarrange1')
        self.assertEqual(df['experiment_name'][random_row(TASK_ROWS)], 'myExp')
        self.assertEqual(df['version'][random_row(TASK_ROWS)], '2')
        ## trial-level descriptors
        TRIAL = 23
        TRIAL_ROWS = 4
        trial_df = df[df['trial'] == TRIAL].reset_index()
        self.assertEqual(len(trial_df), TRIAL_ROWS)
        self.assertEqual(trial_df['trial_start'][random_row(TRIAL_ROWS)], 1646494875092)
        self.assertEqual(trial_df['trial_end'][random_row(TRIAL_ROWS)], 1646494884936)
        self.assertEqual(trial_df['n_actions'][random_row(TRIAL_ROWS)], 4)
        ## stimulus-level descriptors
        self.assertEqual(trial_df['stim_id'][0], '6f3bcd859365621cb1fac620122aad98')
        self.assertEqual(trial_df['x'][0], 0.316031215513301)
        self.assertEqual(trial_df['y'][0], -0.690127377517038)
        self.assertEqual(trial_df['stim_name'][0], 'rain')
        self.assertEqual(trial_df['stim_type'][0], 'png')

    def test_df_from_dict(self):
        """Test creating a dataframe from a data dictionary directly.
        """
        from meadows.io.pandas import df_from_task_data
        data = dict()
        data['task'] = dict(name='arrangement')
        data['stimuli'] = [
            dict(id='s1', name='stim1', type='png'),
            dict(id='s2', name='stim2', type='wav'),
            dict(id='s3', name='stim3', type='jpg'),
        ]
        data['trials'] = [
            dict(start=111, end=222, positions=[
                dict(id='s1', x=0.17, y=0.18),
                dict(id='s2', x=0.27, y=0.28),
                dict(id='s3', x=0.37, y=0.38)
            ], log=[
                [111, 'logStarted', {}],
                [112, 'placed', {}],
                [113, 'displaced', {}],
                [114, 'placed', {}],
                [115, 'placed', {}],
                [116, 'finish', {}],
            ]),
            dict(start=333, end=444, positions=[
                dict(id='s2', x=1.27, y=1.28),
                dict(id='s3', x=1.37, y=1.38)
            ], log=[
                [111, 'logStarted', {}],
                [112, 'placed', {}],
                [113, 'displaced', {}],
                [114, 'placed', {}],
                [115, 'displaced', {}],
                [116, 'placed', {}],
                [117, 'finish', {}],
            ]),
        ]
        meta = dict(participant='cuddly-bunny', task_index=3,
            experiment_name='myExp', version='7')
        df = df_from_task_data(data, meta)
        N_PLACEMENTS_TOTAL = TASK_ROWS = 5
        self.assertEqual(len(df), N_PLACEMENTS_TOTAL)
        self.assertEqual(df['participant'][random_row(TASK_ROWS)], 'cuddly-bunny')
        self.assertEqual(df['task_index'][random_row(TASK_ROWS)], 3)
        self.assertEqual(df['task_name'][random_row(TASK_ROWS)], 'arrangement')
        self.assertEqual(df['experiment_name'][random_row(TASK_ROWS)], 'myExp')
        self.assertEqual(df['version'][random_row(TASK_ROWS)], '7')
        ## trial-level descriptors
        TRIAL = 1
        TRIAL_ROWS = 2
        trial_df = df[df['trial'] == TRIAL].reset_index()
        self.assertEqual(len(trial_df), TRIAL_ROWS)
        self.assertEqual(trial_df['trial_start'][random_row(TRIAL_ROWS)], 111)
        self.assertEqual(trial_df['trial_end'][random_row(TRIAL_ROWS)], 117)
        self.assertEqual(trial_df['n_actions'][random_row(TRIAL_ROWS)], 5)
        ## stimulus-level descriptors
        self.assertEqual(trial_df['stim_id'][0], 's2')
        self.assertEqual(trial_df['x'][0], 1.27)
        self.assertEqual(trial_df['y'][0], 1.28)
        self.assertEqual(trial_df['stim_name'][0], 'stim2')
        self.assertEqual(trial_df['stim_type'][0], 'wav')

    def test_load_checks_dataframe_from_data(self):
        """Create dataframe from acceptance checks in ma data.
        """
        from meadows.io.pandas import df_checks_from_task_data
        data = dict(trials=[
            dict(annotations=[
                dict(
                    ids=['5c', '3b', '3b', 'f63'],
                    start='2022-02-21T16:21:33.703Z',
                    resp='2022-02-21T16:21:38.128Z',
                    label='no'
                ),
                dict(
                    ids=['5c', '3b', '3b', 'f63'],
                    start='2022-02-21T16:21:33.703Z',
                    resp='2022-02-21T16:21:38.128Z',
                    label='no'
                ),
            ]),
            dict(annotations=[
                dict(
                    ids=['5c', '3b', '3b', 'f63'],
                    start='2022-02-21T16:21:33.703Z',
                    resp='2022-02-21T16:21:38.128Z',
                    label='no'
                ),
            ]),
        ])
        df = df_checks_from_task_data(data)
        self.assertEqual(len(df), 3)
        self.assertEqual(df.iloc[0].stim1_id, '5c')
        self.assertEqual(df.iloc[0].stim2_id, '3b')
        self.assertEqual(df.iloc[0].stim3_id, 'f63')
        self.assertEqual(df.iloc[0].start.to_pydatetime(),
            datetime(2022, 2, 21, 16, 21, 33, 703000, tzinfo=tzutc()))
        self.assertEqual(df.iloc[0].resp.to_pydatetime(),
            datetime(2022, 2, 21, 16, 21, 38, 128000, tzinfo=tzutc())) 
        self.assertEqual(df.iloc[0].label, 'no')



import unittest
from numpy.testing import assert_array_almost_equal, assert_allclose
import numpy, random


class RandomNumberGeneratorTests(unittest.TestCase):

    def test_jsf32_randomgen(self):
        from meadows.rng import RandomNumberGenerator
        rng = RandomNumberGenerator(engine='randomgen')
        assert_array_almost_equal(rng.random(4, 0), [
                0.10393405123613775,
                0.6028600085992366,
                0.9420762336812913,
                0.035197859862819314,
            ],
            decimal=12
        )

    def test_jsf32_pure_python(self):
        from meadows.rng import RandomNumberGenerator
        rng = RandomNumberGenerator(engine='python')
        assert_array_almost_equal(rng.random(4, 0), [
                0.10393405123613775,
                0.6028600085992366,
                0.9420762336812913,
                0.035197859862819314,
            ],
            decimal=12
        )

    def test_sample_indices_fixed_seed(self):
        from meadows.rng import RandomNumberGenerator
        rng = RandomNumberGenerator(engine='randomgen')
        idx = rng.sample_indices(pop_size=1000, n=7, seed=0)
        self.assertEqual(idx, [103, 603, 942, 35, 130, 512, 47])

    def test_sample_indices_will_not_sample_more_than_pop_size(self):
        from meadows.rng import RandomNumberGenerator
        rng = RandomNumberGenerator()
        idx = rng.sample_indices(pop_size=10, n=12, seed=0)
        self.assertEqual(len(idx), 10)

    def test_sample_indices_is_uniform(self):
        from meadows.rng import RandomNumberGenerator
        rng = RandomNumberGenerator(engine='randomgen')
        chosen_count = numpy.zeros([10,])
        s = 100*1000
        for _ in range(s):
            seed = random.randint(294967296, 4294967296)
            idx = rng.sample_indices(pop_size=10, n=5, seed=seed)
            chosen_count[idx] += 1
        assert_allclose(
            chosen_count,
            numpy.ones([10,])*(s/2),
            rtol=0.01
        )

    def test_sample_indices_are_always_unique(self):
        from meadows.rng import RandomNumberGenerator
        rng = RandomNumberGenerator(engine='randomgen')
        s = 1000
        for _ in range(s):
            seed = random.randint(294967296, 4294967296)
            idx = rng.sample_indices(pop_size=10, n=5, seed=seed)
            self.assertEqual(len(idx), len(set(idx)))



"""Tools related to the Multiple Arrangements task
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Tuple
from warnings import warn
import numpy
from pandas import DataFrame
from scipy.stats import spearmanr
from sklearn.manifold import MDS
from rsatoolbox.rdm.calc import calc_rdm_euclid
from rsatoolbox.rdm.combine import from_partials, rescale
from rsatoolbox.data import Dataset
from meadows.io.pandas import df_checks_from_task_data, df_from_task_data


def calc_trial_rep(ds: Dataset) -> float:
    ## TODO: arg: include_partial_trials
    ## can use smacoff with nans or 0s?
    ## or median distance
    """Calculate MA Trial Replicability Index

    Args:
        ds (Dataset): rsatoolbox Dataset with the MA experiment data

    Returns:
        float: replicability index
    """
    mds = MDS(
        n_components=2,
        random_state=numpy.random.RandomState(seed=1),
        dissimilarity='precomputed'
    )
    rdm_list = []
    for ds in ds.split_obs('trial'):
        rdm = calc_rdm_euclid(ds)
        rdm.pattern_descriptors['stim_name'] = ds.obs_descriptors['stim_name']
        rdm_list.append(rdm)
    n_trials = len(rdm_list)
    trial_wise_prediction = numpy.full(n_trials, numpy.nan)
    for t in range(n_trials):
        test_rdm = rdm_list[t]
        test_items = test_rdm.pattern_descriptors['stim_name']
        training_rdm_list = [rdm for r, rdm in enumerate(rdm_list) if r != t]
        training_rdms = from_partials(training_rdm_list, descriptor='stim_name')
        training_rdms_scaled = rescale(training_rdms, method='evidence')
        training_rdm = training_rdms_scaled.mean(weights='rescalingWeights')
        training_subset = training_rdm.subset_pattern('stim_name', test_items)
        n_nans = numpy.sum(numpy.isnan(training_subset.dissimilarities))
        if n_nans > 0:
            warn(f'Training trials missing {n_nans} pairs, skipping fold.')
            continue
        predicted = mds.fit_transform(training_subset.get_matrices()[0, :, :])
        predicted_rdm = calc_rdm_euclid(Dataset(predicted))
        predicted_rdm.pattern_descriptors['stim_name'] = training_subset.pattern_descriptors['stim_name']
        predicted_rdm.sort_by(stim_name=test_rdm.pattern_descriptors['stim_name'])
        trial_wise_prediction[t], _ = spearmanr(
            predicted_rdm.dissimilarities[0,:], test_rdm.dissimilarities[0,:])
    return numpy.nanmean(trial_wise_prediction)

def evaluate_checks(data: Dict) -> Tuple[float, DataFrame]:
    """Compare attention check responses to arranged distances

    Args:
        data (Dict): MA task data

    Returns:
        Tuple[float, DataFrame]: _description_
    """
    ## Did you place the top pair further apart than the bottom pair?
    df_ma = df_from_task_data(data)
    stim_set = set([(r.stim_id, r.stim_name) for _, r in df_ma.iterrows()])
    stim_dict = dict(list(stim_set))
    df = df_checks_from_task_data(data)
    triplets = 'stim2_id' in df
    stim_nrs = (1, 2, 3) if triplets else (1, 2)
    pairs = [(1, 2), (2, 3)] if triplets else [(1, 2)]
    for nr in stim_nrs:
        names = [stim_dict[i] for i in df[f'stim{nr}_id']]
        df[f'stim{nr}_name'] = names
    for r, row in df.iterrows():
        for p, pair in enumerate(pairs, start=1):
            pair_pos = []
            for pair_index in pair:
                sid = row[f'stim{pair_index}_id']
                pos = df_ma.loc[(df_ma.trial == row.trial) & (df_ma.stim_id == sid)]
                pair_pos.append(numpy.array([pos.x, pos.y]))
            df.loc[r, f'pair{p}_dist'] = numpy.linalg.norm(pair_pos[0]-pair_pos[1])
    ## below analysis part specific for "repeat" mode with triplets
    df['true'] = df.pair1_dist > df.pair2_dist
    df['acc'] = df.true == (df.label == 'yes')
    ## remove ids for readability
    df = df[[c for c in df.columns if '_id' not in c]]
    return (df.acc.mean(), df)



from __future__ import annotations
from pandas import DataFrame
import numpy
from dateutil.parser import isoparse as parse_iso_dt


def score_cfpt(annotations: DataFrame) -> DataFrame:
    """Trial-wise score and RT for a CFPT OrderSort task annotations

    To get a dictionary of total scores by participant:
    ```
        df.groupby('participation').sum().score.to_dict()
    ```

    Args:
        annotations (DataFrame): Annotations

    Returns:
        Tuple[Dict[str, int], DataFrame]: First element is the
            dictionary of total scores,
            second a copy of the dataframe with a new "score" column
    """
    rows = []
    for _, trial in annotations.iterrows():
        start_dt = parse_iso_dt(trial.start)
        resp_dt = parse_iso_dt(trial.resp)
        n_moves = int(trial.label.split('_')[0])
        stims_str = '_'.join(trial.label.split('_')[1:])
        parts = dict()
        for p, part in enumerate(['cond', 'id', 'mix']):
            parts[part] = [s.split('_')[p] for s in stims_str.split('-')]
        score = numpy.abs(numpy.argsort(parts['mix']) - numpy.arange(6)).sum()
        rows.append(dict(
            participation=trial.participation,
            identity=int(parts['id'][0]),
            condition=parts['cond'][0],
            n_moves=n_moves,
            rt=(resp_dt-start_dt).total_seconds(),
            score=score
        ))
    return DataFrame(rows)
    


from __future__ import annotations
from typing import Dict, Optional
import json
from pandas import DataFrame
from dateutil.parser import isoparse as parse_iso_dt
from rsatoolbox.io.meadows import extract_filename_segments


def df_from_tree_file(fpath: str) -> DataFrame:
    """Read data from json file and return as pandas DataFrame

    Currently only supports single task, single participant MA 

    Args:
        fpath (str): full path to json file

    Returns:
        DataFrame: long-form DataFrame
    """
    meta = extract_filename_segments(fpath)
    with open(fpath) as fhandle:
        data = json.load(fhandle)
    return df_from_task_data(data, meta)

def df_from_task_data(data: dict, meta: Optional[Dict]=None) -> DataFrame:
    """Create a pandas DataFrame for MultiArrange task data

    Args:
        data (dict): Data for one MultiArrangement task
        meta (Optional[Dict], optional): Optional dictionary with metadata.
            Defaults to None.

    Returns:
        DataFrame: rows are item placements, with xy coordinates, trial index
            and more as columns.
    """
    if meta is None:
        meta = dict()
    stim_names = dict([(s['id'], s['name']) for s in data['stimuli']])
    stim_types = dict([(s['id'], s['type']) for s in data['stimuli']])
    rows = []
    for t, trial in enumerate(data['trials']):
        actions = [e for e in trial['log'] if e[1] in ('placed', 'displaced')]
        start_event = [e for e in trial['log'] if e[1] == 'logStarted'][0]
        finish_event = [e for e in trial['log'] if e[1] == 'finish'][0]
        for position in trial['positions']:
            rows.append(dict(
                stim_id=position['id'],
                stim_name=stim_names[position['id']],
                stim_type=stim_types[position['id']],
                x=position['x'],
                y=position['y'],
                trial=t,
                n_actions=len(actions),
                trial_start=start_event[0],
                trial_end=finish_event[0],
            ))
        
    df = DataFrame(rows)
    df['task_name'] = data['task']['name']
    for key in ['participant', 'task_index', 'experiment_name', 'version']:
        if key in meta:
            df[key] = meta[key]
    return df

def df_checks_from_task_data(data: dict) -> DataFrame:
    """Create a dataframe with attention checks from an MA task

    Args:
        data (dict): data for one task such as loaded from json

    Returns:
        DataFrame: Each row is one check
    """
    rows = []
    for t, trial in enumerate(data['trials']):
        for annotation in trial['annotations']:
            row = dict(
                trial=t,
                stim1_id=annotation['ids'][0],
                stim2_id=annotation['ids'][1],
                label=annotation['label'],
                start=parse_iso_dt(annotation['start']),
                resp=parse_iso_dt(annotation['resp']),
            )
            if len(annotation['ids']) > 2:
                row['stim3_id'] = annotation['ids'][3]
            rows.append(row)
    return DataFrame(rows)



"""RandomNumberGenerator class

Requires:
    randomgen==1.19.3
"""
import numpy
from randomgen import JSF, SeedSequence
from meadows.jsf32 import JSF32


class RandomNumberGenerator(object):
    """Custom PRNG which is guaranteed to produce the same series
    as the Typescript equivalent, given the same seed.

    Currently only implements JSF32, as this is the first one I found with
    a python implementation that I could get to produce the same numbers.
    A seed is expected to be provided with each call.

    Provides two implementations, one using the RandomGen package,
    which is in the process of being integrated into numpy.
    And a pure-python one which is ~1000x slower, but remains here in
    case the other implementation happens to dissapear.
    """

    def __init__(self, engine='randomgen'):
        """[summary]

        Args:
            engine (str, optional): 'python' for the slow, pure-python
                version or 'randomgen', for the c-based, fast implementation
                using the RandomGen package. Defaults to 'randomgen'.
        """
        self.engine = engine

    def random(self, n=1, seed=0):
        """Generate a list of floats

        4294967296 = 2**32
        4058668781 = 0xF1EA5EED

        Args:
            n (int, optional): How many numbers to generate. Defaults to 1.
            seed (int, optional): The seed for the bitgenerator.
                Should be anywhere from 0 to 2**32. Defaults to 0.

        Returns:
            list: List of floats between 0 and 1.
        """
        if self.engine == 'randomgen':
            ss = DirectSeedSequence(seed)
            jsf = JSF(seed=ss, size=32)
            return [jsf.random_raw()/4294967296 for _ in range(n)]
        elif self.engine == 'python':
            jsf = JSF32(4058668781, seed, seed, seed)
            _ = [jsf.next() for _ in range(20)]
            return [jsf.next() for _ in range(n)]

    def sample_indices(self, pop_size=10, n=5, seed=0):
        """Choose a number of integers between 0 and POP_SIZE, without
        replacement.

        Args:
            pop_size (int, optional): Size of the population. Defaults to 10.
            n (int, optional): How many indices to pick. Defaults to 5.
            seed (int, optional): Seed for the bitgenerator.
                Should be anywhere from 0 to 2**32. Defaults to 0.

        Returns:
            list: List of integers
        """
        n = min(n, pop_size)
        random_floats = self.random(n=n, seed=seed)
        sample = []
        population = list(range(pop_size))
        for r in range(n):
            i = int(random_floats[r] * len(population))
            sample.append(population.pop(i))
        return sample


class DirectSeedSequence(SeedSequence):
    """A simple seed sequence mock which simply passes the single integer.

    See RandomGen / numpy documentation
    """

    def __init__(self, seed_int):
        self.seed_int = seed_int

    def generate_state(self, n_words, dtype):
        return numpy.array([self.seed_int] * n_words, dtype=dtype)





def rshift(val, n):
    """Zero-fill right shift bit-wise operator

    Equivalent of >>> in javascript

    Args:
        val ([type]): [description]
        n ([type]): [description]

    Returns:
        [type]: [description]
    """
    return (val % 0x100000000) >> n

class JSF32(object):
    """Pure-python implementation of JSF32

    Takes 1.22 ms

    Usage:
        gen = JSF32(4058668781, 0, 0, 0) ## aka 0xF1EA5EED
        print(gen.next())
        print(gen.next())
    """
    def __init__(self, a,b,c,d):
        self.a =a
        self.b = b
        self.c = c
        self.d = d

    def next(self):
        self.a = self.a | 0
        self.b = self.b |  0
        self.c = self.c |  0
        self.d = self.d |  0
        t = self.a - (self.b << 27 | rshift(self.b,5) ) | 0
        self.a = self.b ^ (self.c << 17 | rshift(self.c,15) )
        self.b = self.c + self.d | 0
        self.c = self.d + t | 0
        self.d = self.a + t | 0
        return rshift(self.d, 0) / 4294967296



"""Tests for the MA attention checks
"""
#pylint: disable=import-outside-toplevel, no-self-use
from unittest import TestCase
import pkg_resources, json


class MaChecksTests(TestCase):
    """Tests for the MA attention checks
    """

    def test_repeat_unit(self):
        pass

    def test_from_1p_1t_json(self):
        """Test on sample data
        """
        from meadows.multiarrange import evaluate_checks
        fname = 'Meadows_MaChecks_v_vdesign_preview_1_tree.json'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        with open(fpath) as fhandle:
            data = json.load(fhandle)
        acc, df = evaluate_checks(data)
        self.assertEqual(acc, 1.0)
        self.assertAlmostEqual(df.iloc[0].pair1_dist, 0.6321, 4)
        self.assertAlmostEqual(df.iloc[0].pair2_dist, 0.8214, 4)
        self.assertEqual(df.iloc[0].label, 'no')
        self.assertEqual(df.iloc[0].acc, True)



from unittest import TestCase
import pkg_resources
from numpy import array
from numpy.testing import assert_array_almost_equal


class IoRsatoolboxTests(TestCase):
    """Acceptance and unit tests for loading DataSets from json files.
    """

    def test_load_datasets_from_json_file_1p_1t(self):
        """Acceptance test for loading data from a Meadows
        .json file download containing data for a single MA task,
        single participant. Should have data and descriptors
        as found in file.
        """
        from meadows.io.rsatoolbox import load_dataset
        fname = 'Meadows_myExp_v_v2_profound-mammoth_1_tree.json'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        ds = load_dataset(fpath)
        self.assertEqual(ds.descriptors.get('participant'), 'profound-mammoth')
        self.assertEqual(ds.descriptors.get('task_index'), 1)
        self.assertEqual(ds.descriptors.get('task_name'), 'multiarrange1')
        self.assertEqual(ds.descriptors.get('experiment_name'), 'myExp')
        self.assertEqual(ds.descriptors.get('version'), '2')
        self.assertEqual(ds.channel_descriptors.get('name'), ['x', 'y'])
        ## obs
        O = 99
        self.assertEqual(ds.obs_descriptors.get('trial')[O], 23)
        self.assertEqual(ds.obs_descriptors.get('trial_start')[O], 1646494875092)
        self.assertEqual(ds.obs_descriptors.get('trial_end')[O], 1646494884936)
        self.assertEqual(ds.obs_descriptors.get('n_actions')[O], 4)
        self.assertEqual(ds.obs_descriptors.get('stim_id')[O], 'fc3317dadd5948d807175da02bc977c4')
        self.assertEqual(ds.obs_descriptors.get('stim_name')[O], 'mud_bubbling')
        ## overall descriptor since the same for all stims:
        self.assertEqual(ds.descriptors.get('stim_type'), 'png')
        assert_array_almost_equal(ds.measurements[O, :], array([ 0.18029752, -0.94403202]))



"""Unit tests for 
"""
#pylint: disable=import-outside-toplevel
from unittest import TestCase
import numpy
from numpy.testing import assert_array_equal
import pandas


class CfmtScoreTests(TestCase):
    """Unit tests for score_cfmt
    """

    def test_score(self):
        """score_cfmt() adds the "correct" boolean column and a total score
        """
        from meadows.cfmt import score_cfmt
        annotations = pandas.DataFrame([
            ['pract_00_X_01_13_l_li', 1],
            ['pract_00_X_02_00_f_li', 2],
            ['pract_00_X_03_13_r_li', 3], # wrong
            ['intro_00_B_01_13_l_li', 2],
            ['intro_00_B_02_00_f_li', 1], # wrong
            ['intro_00_B_03_13_r_li', 3],
            ['intro_01_D_01_13_l_li', 2],
            ['intro_01_D_02_00_f_li', 3],
            ['intro_05_Z_02_00_f_li', 3],
            ['intro_05_Z_03_13_r_li', 2], # wrong
            ['novel_00_M_05_00_f_hl', 3], # wrong
            ['novel_01_N_08_13_r_dk', 2],
            ['novel_28_N_06_23_l_li', 1],
            ['novel_29_B_05_00_f_hl', 3],
            ['noise_00_Z_12_00_f_li', 2],
            ['noise_01_M_11_23_r_li', 2], # wrong
            ['noise_22_B_11_23_r_li', 1],
            ['noise_23_D_11_23_r_li', 2]  # wrong
        ], columns=['stim1_name', 'label'])
        out_score, out_df = score_cfmt(annotations)
        self.assertEqual(out_score, 10)
        assert_array_equal(out_df.correct.values, numpy.array([
            True, False, True, True, True, True,
            False, False, True, True, True, True, False, True, False
        ]))



from rsatoolbox.data.dataset import Dataset
from meadows.io.pandas import df_from_tree_file


def load_dataset(fpath: str) -> Dataset:
    df = df_from_tree_file(fpath)
    return Dataset.from_df(df)



"""Tests for the trial replicability index
"""
#pylint: disable=import-outside-toplevel, no-self-use
from unittest import TestCase
import pkg_resources
from numpy.random import RandomState
import numpy
from rsatoolbox.data import Dataset


class MaTrialReplicabilityTests(TestCase):
    """Tests for the trial replicability index
    """

    def test_index_high(self):
        from meadows.multiarrange import calc_trial_rep
        ds = Dataset(
            measurements=numpy.array([
                [-1,    0   ], # trial 1, courgette
                [-0.9,  0.1 ], # trial 1, cucumber
                [ 0.4,  0.4 ], # trial 1, banana
                [ 0.7,  0.7 ], # trial 1, apple
                [ 0.6,  0.75], # trial 1, orange
                [ 0,   -1   ], # trial 2, banana
                [ 0.2,  0.9 ], # trial 2, apple
                [-0.2,  0.9 ], # trial 2, orange
                [ 0,    -1  ], # trial 3, courgette
                [-0.1,  0.8 ], # trial 3, cucumber
                [ 0,    1   ], # trial 3, banana
            ]),
            descriptors=dict(foo='bar'),
            obs_descriptors=dict(
                trial=[1,1,1,1,1,2,2,2,3,3,3],
                stim_name=['co', 'cu', 'ba', 'ap', 'or', 'ba', 'ap', 'or', 
                    'co', 'cu', 'ba']
            ),
            channel_descriptors=dict(dimension=['x', 'y'])
        )
        self.assertAlmostEqual(calc_trial_rep(ds), 0.683, places=3)

    def test_index_random(self):
        from meadows.multiarrange import calc_trial_rep
        rng = RandomState(seed=1)
        ds = Dataset(
            measurements=(rng.rand(11, 2)*2)-1,
            descriptors=dict(foo='bar'),
            obs_descriptors=dict(
                trial=[1,1,1,1,1,2,2,2,3,3,3],
                stim_name=['co', 'cu', 'ba', 'ap', 'or', 'ba', 'ap', 'or', 
                    'co', 'cu', 'ba']
            ),
            channel_descriptors=dict(dimension=['x', 'y'])
        )
        self.assertAlmostEqual(calc_trial_rep(ds), 0.0)

    def test_load_datasets_from_json_file_1p_1t(self):
        """Test on sample data
        """
        from meadows.multiarrange import calc_trial_rep
        from meadows.io.rsatoolbox import load_dataset
        fname = 'Meadows_myExp_v_v2_profound-mammoth_1_tree.json'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        ds = load_dataset(fpath)
        self.assertAlmostEqual(calc_trial_rep(ds), 0.9057, places=4)



from setuptools import setup, find_packages

requires = []
with open('requirements.txt') as reqfile:
    requires = reqfile.read().splitlines()

setup(
    name='meadows',
    version='0.0.4',
    description='Meadows Scythe: the Meadows Export & Analysis toolkit',
    url='https://github.com/meadowsresearch/scythe',
    long_description='Library for meadows users to preprocess downloaded data',
    long_description_content_type='text/plain',
    classifiers=[
      'Programming Language :: Python',
      'Topic :: Internet :: WWW/HTTP',
      'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
      'Intended Audience :: Science/Research',
      'Programming Language :: Python :: 3.7',
      'Programming Language :: Python :: 3.8',
      'Programming Language :: Python :: 3.9',
    ],
    author='Jasper van den Bosch',
    author_email='jasper@meadows-research.com',
    keywords='',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    tests_require=requires,
    test_suite='tests',
)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from os.path import join, splitext, isdir
from os import mkdir
from tempfile import TemporaryDirectory
from matplotlib.pyplot import close, subplots, text
from matplotlib.patches import Circle
from imageio import get_writer as get_iio_writer, imread
if TYPE_CHECKING:
    from rsatoolbox.data.dataset import Dataset
    from matplotlib.axes._axes import Axes
    from matplotlib.figure import Figure


def arrangement(ds: Dataset, media_path: str, item_size=5.0, ax: Optional[Axes]=None) -> Optional[Figure]:
    """Plot a single trial of a MultiArrangement task

    Args:
        ds (Dataset): rsatoolbox Dataset containing a single trial
        media_path (str): Path to directory containing your stimuli.
        item_size (int): The stimulus size as you configured it on Meadows.
            This is the height of the stimulus in % of the width of the window,
            (note the height of the window is 2x the width). Default is 5.
        ax (Optional[Axes], optional): Matplotlib axes to plot the trial on.
            Defaults to None.

    Raises:
        ValueError: Will throw if Dataset contains multiple trials

    Returns:
        Optional[Figure]: If no axes passed, will create a figure and return it.
    """

    half_heigth = 1/(1 - (1.25 * 0.04 * item_size)) # half axis height in data pts
    item_extent = half_heigth * 0.04 * item_size    # item height in data pts
    
    trial_idx = ds.obs_descriptors['trial']
    if len(set(trial_idx)) > 1:
        raise ValueError('DataSet contains multiple trials, expecting just one')
    fig = None
    if ax is None:
        fig, ax = subplots(figsize=(8, 8))

    ax.set_xlim([-half_heigth, half_heigth])
    ax.set_ylim([-half_heigth, half_heigth])
    arena = Circle((0, 0), 1, color='w')
    ax.add_patch(arena)
    ax.set_facecolor((0.866, 0.866, 0.866)) # #DDD, meadows bg color
    ax.axes.xaxis.set_visible(False)
    ax.axes.yaxis.set_visible(False)
    ax.set_title(f'trial {trial_idx[0]+1}')

    coords = ds.get_measurements()
    fnames = ds.obs_descriptors['stim_name']
    if 'stim_type' in ds.descriptors:
        stim_types = [ds.descriptors['stim_type']]*len(fnames)
    else:
        stim_types = ds.obs_descriptors['stim_type']
    for i in range(ds.n_obs):
        fpath = join(media_path, f'{fnames[i]}.{stim_types[i]}')
        img = imread(fpath)
        x, y = coords[i, :]
        y = -y ## flip given meadows coords vs mpl coords
        ax.imshow(img, zorder=3, extent=(
            x-(item_extent/2),
            x+(item_extent/2),
            y-(item_extent/2),
            y+(item_extent/2),
        )) #(left, right, bottom, top)

    ## add right-side metadata annotations
    trial_start = ds.obs_descriptors['trial_start'][0]
    trial_end = ds.obs_descriptors['trial_end'][0]
    n_actions = ds.obs_descriptors['n_actions'][0]
    dur_secs = (trial_end-trial_start)/1000
    trial_meta = [
        ('items', ds.n_obs, ''),
        ('time/item', dur_secs/ds.n_obs, 's'),
        ('actions/item', n_actions/ds.n_obs, ''),
    ]
    text_kwargs = dict(horizontalalignment='right', verticalalignment='center')
    y_offset = 0.1
    for label, val, suffix in trial_meta:
        content = f'{label}: {round(val, 1)}{suffix} '
        text(half_heigth, half_heigth-y_offset, content, **text_kwargs)
        y_offset += 0.1
    return fig


def arrangements(ds: Dataset, fpath='.', media_path: str='', item_size=5.0):
    """Plot each trial in a MultiArrange Dataset

    Args:
        ds (Dataset): rsatoolbox Dataset for a MultiArrangement task.
            See meadows.io.rsatoolbox.load_dataset()
        fpath (str, optional): Where to save the plots. The format, either a gif
            or a directory with pngs, will be inferred from the extension.
            Defaults to '.', saving pngs in the current directory.
        media_path (str, optional): Path to directory containing your stimuli.
            Defaults to ''.
        item_size (int): The stimulus size as you configured it on Meadows.
            This is the height of the stimulus in % of the width of the window,
            (note the height of the window is 2x the width). Default is 5.
    """
    trial_ds_list = ds.split_obs('trial')
    with create_directory(fpath) as dir_path:
        with get_writer(fpath, mode='I', duration=1) as writer:
            for trial_ds in trial_ds_list:
                fig = arrangement(trial_ds, media_path=media_path, item_size=item_size)
                assert fig is not None
                trial_idx = trial_ds.obs_descriptors['trial']
                fpath = join(dir_path, f'trial_{trial_idx[0]+1}.png')
                fig.savefig(fpath)
                close(fig)
                writer.append_data(imread(fpath))


def is_directory_path(fpath: str) -> bool:
    _, ext = splitext(fpath)
    return ext not in ('.gif',)


def create_directory(fpath: str):
    if is_directory_path(fpath):
        return PermanentDirectory(fpath)
    else:
        return TemporaryDirectory()


def get_writer(fpath: str, mode: str, duration: int):
    if is_directory_path(fpath):
        return DummyWriter()
    else:
        return get_iio_writer(fpath, mode=mode, duration=duration)


class PermanentDirectory(object):
    def __init__(self, dir_path: str):
        self.dir_path = dir_path
    def __enter__(self):
        if not isdir(self.dir_path):
            mkdir(self.dir_path)
        return self.dir_path
    def __exit__(self, type, value, traceback):
        pass

class DummyWriter(object):
    def __init__(self, *args, **kwargs):
        pass
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        pass
    def append_data(self, img):
        pass
